<?php

namespace Drupal\openid_connect_akamai\EventSubscriber;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Url;
use Drupal\openid_connect\Plugin\OpenIDConnectClientManager;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\RequestException;
use Symfony\Component\HttpKernel\KernelEvents;
use Symfony\Component\HttpKernel\Event\GetResponseEvent;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Force redirect anonymous.
 *
 * Force redirect anonymous user requests from site
 * to Akamai login endpoint for SSO.
 */
class AnonymousRedirectSubscriber implements EventSubscriberInterface {

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * OpenId Akamai Client.
   *
   * @var \Drupal\openid_connect_akamai\Plugin\OpenIDConnectClient\Akamai
   */
  protected $openidAkamaiClient;

  /**
   * The OpenId Client Akamai settings.
   *
   * @var array
   */
  protected $akamaiSettings;

  /**
   * Guzzle Http Client.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * Constructs a new AnonymousRedirectSubscriber instance.
   *
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *   The current user.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config
   *   Configuration object.
   * @param \Drupal\openid_connect\Plugin\OpenIDConnectClientManager $openid_client_manager
   *   OpenId Client Manager.
   * @param \GuzzleHttp\ClientInterface $http_client
   *   Guzzle Http Client.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function __construct(AccountInterface $current_user, ConfigFactoryInterface $config, OpenIDConnectClientManager $openid_client_manager, ClientInterface $http_client) {
    $this->currentUser = $current_user;
    $this->akamaiSettings = $config->get('openid_connect.settings.akamai')->get('settings');
    $this->openidAkamaiClient = $openid_client_manager->createInstance('akamai', $this->akamaiSettings);
    $this->httpClient = $http_client;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    $events[KernelEvents::REQUEST][] = ['redirectAnonymous', 100];

    return $events;
  }

  /**
   * Redirects anonymous users to login endpoint for SSO.
   *
   * @param \Symfony\Component\HttpKernel\Event\GetResponseEvent $event
   *   The request event.
   */
  public function redirectAnonymous(GetResponseEvent $event) {

    if ($this->akamaiSettings['enable_sso'] && $this->currentUser->isAnonymous()) {
      $request = $event->getRequest();

      $route_name = Url::fromUserInput($request->getPathInfo())->getRouteName();
      $white_list = ['openid_connect.redirect_controller_redirect'];

      if ($request->cookies->has('akamai_error')
        || in_array($route_name, $white_list)
        || substr($route_name, 0, 5) == 'user.'
      ) {
        return;
      }

      $uri = trim($this->akamaiSettings['domain_endpoint'], '/ ') . '/';
      $uri .= $this->akamaiSettings['customer_id'];
      $uri .= '/login/.well-known/openid-configuration';
      try {
        $this->httpClient->get($uri);
        // Expected result.
        // getBody() returns an instance of Psr\Http\Message\StreamInterface.
        // @see http://docs.guzzlephp.org/en/latest/psr7.html#body
        $response = $this->openidAkamaiClient->authorize();
        $response->setTrustedTargetUrl($response->getTargetUrl() . '&prompt=none');
        $event->setResponse($response);

        $_SESSION['openid_connect_destination'] = $request->getRequestUri();
      }
      catch (RequestException $e) {
      }
    }

  }

}
